#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Provides useful utilities for the other modules as well as for general use.
"""

import lxml
from lxml import etree
from bs4 import BeautifulSoup
from pathlib import Path
from datetime import datetime
import hashlib, uuid, base58, requests, tqdm
import zipfile

XML_NS = "{http://www.w3.org/XML/1998/namespace}"

def elem_iter(root, ns="{http://www.tei-c.org/ns/1.0}"):
    """
    Return an iterator of the elements (utterances, notes, segs, pbs) in a protocol body
    """
    for body in root.findall(".//" + ns + "body"):
        for div in body.findall(ns + "div"):
            for ix, elem in enumerate(div):
                if elem.tag == ns + "u":
                    yield "u", elem
                elif elem.tag == ns + "note":
                    yield "note", elem
                elif elem.tag == ns + "pb":
                    yield "pb", elem
                elif elem.tag == ns + "seg":
                    yield "seg", elem
                elif elem.tag == "u":
                    elem.tag = ns + "u"
                    yield "u", elem
                else:
                    print(elem.tag)
                    yield None


def infer_metadata(filename):
    """
    Heuristically infer metadata from a protocol id or filename.

    Returns a dict with keys "protocol", "chamber", "year", and "number"
    """
    metadata = dict()
    filename = filename.replace("-", "_")
    metadata["protocol"] = filename.split("/")[-1].split(".")[0]
    split = filename.split("/")[-1].split("_")
    metadata["document_type"] = split[0]

    # Year
    for s in split:
        yearstr = s[:4]
        if yearstr.isdigit():
            year = int(yearstr)
            if year > 1800 and year < 2100:
                metadata["year"] = year
                metadata["sitting"] = str(year)

                # Protocol ids of format 197879 have two years, eg. 1978 and 1979
                if s[4:6].isdigit():
                    metadata["secondary_year"] = year + 1
                    metadata["sitting"] += f"/{s[4:6]}"

    # Chamber
    metadata["chamber"] = "Enkammarriksdagen"
    if "_ak_" in filename:
        metadata["chamber"] = "Andra kammaren"
    elif "_fk_" in filename:
        metadata["chamber"] = "Första kammaren"

    try:
        metadata["number"] = int(split[-1])
    except:
        pass  # print("Number parsing unsuccesful", filename)

    return metadata


def clean_html(s):
    """
    Read a HTML file and turn it into valid XML
    """
    soup = BeautifulSoup(s)
    pretty_html = soup.prettify()
    return etree.fromstring(pretty_html)


def validate_xml_schema(xml_path, schema_path):
    """
    Validate an XML file against a schema.
    """
    xml_file = lxml.etree.parse(xml_path)
    xml_file.xinclude()

    schema = lxml.etree.XMLSchema(file=schema_path)
    is_valid = schema.validate(xml_file)

    return is_valid


def protocol_iterators(corpus_root, start=None, end=None):
    """
    Returns an iterator of protocol paths in a corpus.
    """
    folder = Path(corpus_root)
    for protocol in sorted(folder.glob("**/*.xml")):
        path = protocol.relative_to(".")
        assert (start is None) == (
            end is None
        ), "Provide both start and end year or neither"
        if start is not None and end is not None:
            metadata = infer_metadata(protocol.name)
            year = metadata["year"]
            secondary_year = metadata.get("secondary_year", year)
            if start <= year and end >= secondary_year:
                yield str(protocol.relative_to("."))

        else:
            yield str(protocol.relative_to("."))

def parse_date(s):
    """
    Parse datetimes with special error handling
    """
    try:
        return datetime.strptime(s, "%Y-%m-%d")

    except ValueError:
        if len(s) == 4:
            if int(s) > 1689 and int(s) < 2261:
                return datetime(int(s), 6, 15)
            else:
                return None
        else:
            return None

def get_formatted_uuid(seed=None):
    if seed is None:
        x = uuid.uuid4()
    else:
        m = hashlib.md5()
        m.update(seed.encode('utf-8'))
        x = uuid.UUID(m.hexdigest())

    return f"i-{str(base58.b58encode(x.bytes), 'UTF8')}"


def _download_with_progressbar(url, fname, chunk_size=1024):
    resp = requests.get(url, stream=True)
    total = int(resp.headers.get('content-length', 0))
    with open(fname, 'wb') as file, tqdm.tqdm(
        desc=fname,
        total=total,
        unit='iB',
        unit_scale=True,
        unit_divisor=1024,
    ) as bar:
        for data in resp.iter_content(chunk_size=chunk_size):
            size = file.write(data)
            bar.update(size)

def download_corpus(path="./"):
    p = Path(path)
    url = "https://github.com/welfare-state-analytics/riksdagen-corpus/releases/latest/download/corpus.zip"
    zip_path = p / "corpus.zip"
    corpus_path = p / "corpus"
    if corpus_path.exists():
        print(f"WARNING: data already exists at the path '{corpus_path}'. It will be overwritten once the download is finished.")

    zip_path_str = str(zip_path.relative_to("."))
    extraction_path = str(p.relative_to("."))
    
    # Download file and display progress
    _download_with_progressbar(url, zip_path_str)
    with zipfile.ZipFile(zip_path_str, "r") as zip_ref:
        print("Exract to", corpus_path, "...")
        zip_ref.extractall(extraction_path)

    zip_path.unlink()


from lxml import etree
import re
from pyparlaclarin.read import element_hash
import dateparser
import pandas as pd
from .utils import elem_iter, infer_metadata, parse_date
from .db import load_expressions, filter_db, load_patterns, load_metadata
from .segmentation import (
    detect_mp,
    detect_minister,
    detect_speaker,
    expression_dicts,
    detect_introduction,
    intro_to_dict,
)
from .match_mp import multiple_replace
from datetime import datetime


def redetect_protocol(metadata, protocol):
    tei_ns = ".//{http://www.tei-c.org/ns/1.0}"
    parser = etree.XMLParser(remove_blank_text=True)

    party_mapping, join_intros, mp_db, minister_db, speaker_db = metadata
    join_intros = join_intros[join_intros['protocol'] == protocol]
    
    protocol_id = protocol.split("/")[-1]
    metadata = infer_metadata(protocol)
    root = etree.parse(protocol, parser).getroot()
    
    # Year from the folder name
    year = metadata["year"]
    # Take into account folders such as 198889
    secondary_year = metadata.get("secondary_year", year)

    dates = [
        parse_date(elem.attrib.get("when"))
        for elem in root.findall(tei_ns + "docDate")
        if parse_date(elem.attrib.get("when")).year in [year, secondary_year]
    ]
    
    # Dates from xml is wrong for digitized era        
    if len(dates) > 0:
        start_date, end_date = min(dates), max(dates)          

    else:
        start_date = datetime(year,1,1)
        end_date = datetime(secondary_year,12,31)
    
    year_mp_db = filter_db(mp_db, start_date=start_date, end_date=end_date)
    year_minister_db = filter_db(minister_db, start_date=start_date, end_date=end_date)
    year_speaker_db = filter_db(speaker_db, start_date=start_date, end_date=end_date)
    
    # Introduction patterns
    pattern_db = load_patterns()
    pattern_db = pattern_db[
        (pattern_db["start"] <= year) & (pattern_db["end"] >= year)
    ]

    root, unk = detect_mps(
        root,
        None,
        pattern_db,
        mp_db=year_mp_db,
        minister_db=year_minister_db,
        speaker_db=year_speaker_db,
        metadata=metadata,
        party_map=party_mapping,
        join_intros=join_intros,
        protocol_id=protocol_id,
        unknown_variables=["gender", "party", "other"],
    )

    b = etree.tostring(
        root, pretty_print=True, encoding="utf-8", xml_declaration=True
    )

    f = open(protocol, "wb")
    f.write(b)
    f.close()
    return unk


def detect_mps(root, names_ids, pattern_db, mp_db=None, minister_db=None, minister_db_secondary=None, speaker_db=None, metadata=None, party_map=None, join_intros=None, protocol_id=None, unknown_variables=None):
    """
    Re-detect MPs in a parla clarin protocol, based on the (updated)
    MP database.
    """
    mp_expressions = load_expressions(phase="mp")
    ids_to_join = set(join_intros['xml_id1'].tolist()+join_intros['xml_id2'].tolist())
    
    xml_ns = "{http://www.w3.org/XML/1998/namespace}"
    current_speaker = None
    prev = None
    mp_db_secondary = pd.DataFrame()

    # Extract information of unknown speakers
    unknowns = []

    #print(metadata)
    # For bicameral era, prioritize MPs from the same chamber as the protocol
    if "chamber" in metadata:
        chamber = {'Första kammaren': 1, 'Andra kammaren':2}.get(metadata['chamber'], 0)
        mp_db_secondary = mp_db[mp_db['chamber'] != chamber]
        mp_db = mp_db[mp_db['chamber'] == chamber]
        speaker_db = speaker_db[speaker_db['chamber'] == chamber]

    for tag, elem in elem_iter(root):
        if tag == "u":
            # Deleting and adding attributes changes their order;
            # Mark as 'delete' instead and delete later
            elem.set("prev", "delete")
            elem.set("next", "delete")
            if current_speaker is not None:
                elem.attrib["who"] = current_speaker
                if prev is None:
                    prev = elem
                else:
                    new_prev = prev.attrib[xml_ns + "id"]
                    new_next = elem.attrib[xml_ns + "id"]
                    elem.set("prev", new_prev)
                    prev.set("next", new_next)

            else:
                elem.attrib["who"] = "unknown"
                prev = None
        elif tag == "note":
            if elem.attrib.get("type", None) == "speaker":
                text = elem.text
                # Join split intros detected by BERT
                if elem.attrib.get(xml_ns + "id") in ids_to_join:
                    join_intro = join_intros.loc[
                                (join_intros['xml_id1'] == elem.attrib.get(xml_ns + "id")) |
                                (join_intros['xml_id2'] == elem.attrib.get(xml_ns + "id")), 'text']
                    text = join_intro.iloc[0]

                if type(text) == str:
                    d = intro_to_dict(text, mp_expressions)
                    if 'name' in d:
                        d['name'] = multiple_replace(d['name'])
                        
                    if 'other' in d:
                        # Match minister
                        if 'statsråd' in d["other"].lower() or 'minister' in d["other"].lower():
                            current_speaker = detect_minister(text, minister_db, d)

                        elif 'talman' in d["other"].lower():
                            current_speaker = detect_speaker(text, speaker_db, metadata=metadata)

                        else:
                            current_speaker = None

                    # Match mp if not minister/talman and a name is identified
                    # if current_speaker is None and 'name' in d:
                    elif 'name' in d:

                        current_speaker = detect_mp(d, db=mp_db, party_map=party_map)

                        if current_speaker is None and len(mp_db_secondary) > 0:
                            current_speaker = detect_mp(d, db=mp_db_secondary, party_map=party_map)
                    else:
                        current_speaker = None

                    if current_speaker is None:
                        unknowns.append([protocol_id, elem.attrib.get(f'{xml_ns}id')] + [d.get(key, "") for key in unknown_variables])
                    
                    prev = None

    # Do two loops to preserve attribute order
    for tag, elem in elem_iter(root):
        if tag == "u":
            if elem.attrib.get("prev") == "delete":
                del elem.attrib["prev"]
            if elem.attrib.get("next") == "delete":
                del elem.attrib["next"]

    return root, unknowns


def find_introductions(root, pattern_db, intro_ids, minister_db=None):
    """
    Find instances of curation patterns in all files in a folder.

    Args:
        pattern_db: Patterns to be matched as a Pandas DataFrame.
        folder: Folder of files to be searched.
    """

    # return root
    root.text = None
    current_speaker = None
    expressions, manual = expression_dicts(pattern_db)

    for ix, elem_tuple in enumerate(list(elem_iter(root))):
        tag, elem = elem_tuple
        if tag == "u":
            u = None
            u_parent = elem.getparent()
            u_parent.text = None
            for seg in list(elem):
                if type(seg.text) == str:
                    introduction = detect_introduction(
                        seg, intro_ids
                    )
                    if introduction is not None:
                        pass  # print("NEW", seg.text)
                        seg.tag = "{http://www.tei-c.org/ns/1.0}note"
                        seg.attrib["type"] = "speaker"
                        if u is not None:
                            u.addnext(seg)
                        else:
                            elem.addnext(seg)

                        u = etree.Element("{http://www.tei-c.org/ns/1.0}u")
                        # u.text = None
                        if introduction["who"] is not None:
                            u.attrib["who"] = introduction["who"]
                        else:
                            u.attrib["who"] = "unknown"

                        seg.addnext(u)
                        matched_txt = introduction["txt"]
                        ix = None
                        if matched_txt[-1] != ":" and ":" in seg:
                            ix = len(matched_txt) + seg.text.index(matched_txt)
                        if ":" in matched_txt:
                            ix = matched_txt.index(":")
                            ix = ix + seg.text.index(matched_txt)
                        elif seg.text[-1] != ":" and ":" in seg:
                            ix = seg.text.index(":")

                        if ix is not None:
                            rest = seg.text[ix + 1 :]
                            seg.text = seg.text[: ix + 1]
                            new_seg = etree.SubElement(
                                u, "{http://www.tei-c.org/ns/1.0}seg"
                            )
                            new_seg.text = rest

                    elif u is not None:
                        u.append(seg)
                        u.text = None

        elif tag == "note":
            parent = elem.getparent()
            parent.text = None
            # if not elem.attrib.get("type", None) == "speaker":
            if type(elem.text) == str:

                introduction = detect_introduction(
                    elem, intro_ids
                )

                if introduction is not None:
                    if not elem.attrib.get("type", None) == "speaker":
                        print("NEW note", elem.text)
                        elem.attrib["type"] = "speaker"

                        matched_txt = introduction["txt"]
                        ix = None
                        if matched_txt[-1] != ":" and ":" in elem.text:
                            ix = len(matched_txt) + elem.text.index(matched_txt)
                        if ":" in matched_txt:
                            ix = matched_txt.index(":")
                            ix = ix + elem.text.index(matched_txt)
                        elif elem.text[-1] != ":" and ":" in elem:
                            ix = elem.text.index(":")
                        if ix is not None:
                            rest = elem.text[ix + 1 :].strip()
                            if len(rest) > 0:
                                u = etree.Element("{http://www.tei-c.org/ns/1.0}u")
                                # u.text = None
                                if introduction["who"] is not None:
                                    u.attrib["who"] = introduction["who"]
                                else:
                                    u.attrib["who"] = "unknown"

                                elem.addnext(u)

                                rest = elem.text[ix + 1 :]
                                elem.text = elem.text[: ix + 1]

                                new_seg = etree.SubElement(
                                    u, "{http://www.tei-c.org/ns/1.0}seg"
                                )
                                new_seg.text = rest

                    else:
                        pass  # print("OLD", elem.text)

    return root


def detect_date(root, metadata):
    """
    Detect notes with dates in them. Update docDate metadata accordingly.
    """

    dates = set()
    number_dates = set()
    expression = "\\w{3,5}dagen den (\\d{1,2})\\.? (\\w{3,9}) (\\d{4})"
    expression2 = "\\w{3,5}dagen den (\\d{1,2})\\.? (\\w{3,9})"
    expression3 = "(\\d{1,2})\\.? (\\w{3,9}) (\\d{4,4})"
    protocol_year = metadata["year"]
    protocol_years = {protocol_year, metadata.get("secondary_year", protocol_year)}
    yearless = set()

    for ix, elem_tuple in enumerate(list(elem_iter(root))):
        tag, elem = elem_tuple
        if tag == "note" and type(elem.text) == str and len(" ".join(elem.text.split()))  < 50:
            matches = re.search(expression, elem.text)
            matches2 = re.search(expression2, elem.text)
            matches3 = re.search(expression3, elem.text)

            # Dates with the year included, surely date
            if matches is not None:
                elem.attrib["type"] = "date"
                datestr = matches.group(1) + " " + matches.group(2) + " " + matches.group(3)
                date = dateparser.parse(datestr, languages=["sv"])
                if date is not None:
                    if date.year in protocol_years:
                        number_dates.add(date)

            # Dates with the year included, though unsure if protocol date
            elif matches3 is not None:
                datestr = matches3.group()
                date = dateparser.parse(datestr, languages=["sv"])
                if date is not None:
                    if date.year in protocol_years:
                        dates.add(date)

            # Dates without a year
            elif matches2 is not None:
                elem.attrib["type"] = "date"
                datestr = matches2.group(1) + " " + matches2.group(2)
                yearless.add(datestr)

    if len(dates) > 0:
        protocol_year = list(dates)[0].year
    elif len(number_dates) > 0:
        dates = number_dates
        protocol_year = list(dates)[0].year

    for datestr in yearless:
        date = dateparser.parse(datestr + " " + str(protocol_year), languages=["sv"])
        if date is not None:
            dates.add(date)

    dates = sorted(list(dates))
    tei_ns = "{http://www.tei-c.org/ns/1.0}"
    for text in root.findall(".//" + tei_ns + "text"):
        for front in text.findall(".//" + tei_ns + "front"):

            # Remove old docDates
            for docDate in front.findall(".//" + tei_ns + "docDate"):
                docDate.getparent().remove(docDate)
            for div in front.findall(".//" + tei_ns + "div"):
                for docDate in div.findall(".//" + tei_ns + "docDate"):
                    docDate.getparent().remove(docDate)

            if len(dates) > 0:
                for div in front.findall(".//" + tei_ns + "div"):
                    if div.attrib.get("type") == "preface":
                        for docDate in div.findall(".//" + tei_ns + "docDate"):
                            docDate.getparent().remove(docDate)
                        for date in dates:
                            formatted = date.strftime("%Y-%m-%d")
                            docDate = etree.SubElement(div, "docDate")
                            docDate.text = formatted
                            docDate.attrib["when"] = formatted
            else:
                for div in front.findall(".//" + tei_ns + "div"):
                    if div.attrib.get("type") == "preface":
                        formatted = str(protocol_year)
                        docDate = etree.SubElement(div, "docDate")
                        docDate.text = formatted
                        docDate.attrib["when"] = formatted

    return root, dates


def update_ids(root, protocol_id):
    """
    Update element id's
    """
    ids = set()
    xml_id = "{http://www.w3.org/XML/1998/namespace}id"
    for tag, elem in elem_iter(root):
        if tag == "u":
            if xml_id in elem.attrib:
                ids.add(elem.attrib[xml_id])
        elif xml_id in elem.attrib:
            del elem.attrib[xml_id]

    for tag, elem in elem_iter(root):
        if tag == "u":
            if xml_id not in elem.attrib:
                updated_hash = element_hash(elem, protocol_id)
                hash_ix = 0
                updated_id = "i-" + updated_hash + "-" + str(hash_ix)

                while updated_id in ids:
                    hash_ix += 1
                    updated_id = "i-" + updated_hash + "-" + str(hash_ix)

                elem.attrib[xml_id] = updated_id
                ids.add(updated_id)

            for subelem in elem:
                if xml_id in subelem.attrib:
                    del subelem.attrib[xml_id]

    return root




import pandas as pd
import re
from .match_mp import multiple_replace
from functools import partial
import datetime

def increase_date_precision(date, start=True):
	if pd.isna(date):
		return date
	# Year
	if len(date) == 4 and start:
		return date + '-01-01'
	if len(date) == 4 and not start:
		return date + '-12-31'
	# Month
	if len(date) == 7 and start:
		return date + '-01'
	if len(date) == 7 and start:
		last_day = calendar.monthrange(int(date[0]), int(date[1]))[1]
		return date + f'-{last_day}'
	# Day
	if len(date) == 10:
		return date


def check_date_overlap(start1, end1, start2, end2):
	latest_start = max(start1, start2)
	earliest_end = min(end1, end2)
	delta = (earliest_end - latest_start).days + 1
	overlap = max(0, delta)
	return True if overlap > 0 else False


def impute_member_date(db, gov_db, from_gov='Regeringen Löfven I'):
	gov_start = gov_db.loc[gov_db['government'] == from_gov, 'start'].iloc[0]
	idx = 	(db['source'] == 'member_of_parliament') &\
			(db['start'] > gov_start) &\
			(db['end'].isna())
	db.loc[idx, 'end'] = gov_db['end'].max()
	return db


def impute_minister_date(db, gov_db):
	def _impute_minister_date(minister, gov_db):
		if pd.isna(minister['start']):
			minister['start'] = gov_db.loc[gov_db['government'] == minister['government'], 'start'].iloc[0]
		if pd.isna(minister['end']):
			minister['end'] = gov_db.loc[gov_db['government'] == minister['government'], 'end'].iloc[0]
		return minister

	# Impute missing minister dates using government dates
	db.loc[db['source'] == 'minister'] =\
	db.loc[db['source'] == 'minister'].apply(partial(_impute_minister_date, gov_db=gov_db), axis=1)
	return db


def impute_speaker_date(db):
	idx = 	(db['source'] == 'speaker') &\
			(db['end'].isna()) &\
			(db['role'].str.contains('kammare') == False)
	db.loc[idx, 'end'] = db.loc[idx, 'start'] + datetime.timedelta(days = 365*4)
	return db


def impute_date(db):
	db[["start", "end"]] = db[["start", "end"]].astype(str)
	db['start'] = db['start'].apply(increase_date_precision, start=True)
	db['end'] = db['end'].apply(increase_date_precision, start=False)
	db[["start", "end"]] = db[["start", "end"]].apply(pd.to_datetime, format='%Y-%m-%d')
	
	if 'source' not in db.columns:
		return db

	# Impute current governments end date
	gov_db = pd.read_csv('corpus/metadata/government.csv')
	gov_db[["start", "end"]] = gov_db[["start", "end"]].apply(pd.to_datetime, format='%Y-%m-%d')
	idx = gov_db['start'].idxmax()
	gov_db.loc[idx, 'end'] = gov_db.loc[idx, 'start'] + datetime.timedelta(days = 365*4)

	sources = set(db['source'])
	if 'member_of_parliament' in sources:
		db = impute_member_date(db, gov_db)
	if 'minister' in sources:
		db = impute_minister_date(db, gov_db)
	if 'speaker' in sources:
		db = impute_speaker_date(db)
	return db


def impute_party(db, party):
	if 'party' not in db.columns:
		db['party'] = pd.Series(dtype=str)
	data = []
	for i, row in db[db['party'].isnull()].iterrows():	
		parties = party[party['wiki_id'] == row['wiki_id']]
		if len(set(parties['party'])) == 1:
			db.loc[i,'party'] = parties['party'].iloc[0]
		if len(set(parties['party'])) >= 2:
			for j, sow in parties.iterrows():
				try:
					res = check_date_overlap(row['start'], sow['start'], row['end'], sow['end'])
				except:
					print("Impute dates on Corpus using impute_date() before imputing parties!\n")
					raise
				if res:
					m = row.copy()
					m['party'] = sow['party']
					data.append(m)
	db = pd.concat([db, pd.DataFrame(data)]).reset_index(drop=True)
	return db


def abbreviate_party(db, party):
	party = {row['party']:row['abbreviation'] for _, row in party.iterrows()}
	db["party_abbrev"] = db["party"].fillna('').map(party)
	return db


def clean_name(db):
	idx = db['name'].notna()
	db.loc[idx, 'name'] = db.loc[idx, 'name'].str.lower()
	db.loc[idx, 'name'] = db.loc[idx, 'name'].astype(str).apply(multiple_replace)
	db.loc[idx, 'name'] = db.loc[idx, 'name'].str.replace('-', ' ', regex=False)
	db.loc[idx, 'name'] = db.loc[idx, 'name'].str.replace(r'[^a-zåäö\s\-]', '', regex=True)
	return db


def infer_chamber(db):
	def _infer_chamber(role):
		d = {'första': 1, 'andra': 2}
		match = re.search(r'([a-zåäö]+)\s*(?:kammar)', role)
		return d[match.group(1)] if match else 0
	db['chamber'] = db['role'].apply(_infer_chamber).astype(dtype=pd.Int8Dtype())
	return db


def format_member_role(db):
	db['role'] = db['role'].str.extract(r'(ledamot)')
	return db


def format_minister_role(db):
	db["role"] = db["role"].str.replace('Sveriges ', '').str.lower()
	return db


def format_speaker_role(db):
	def _format_speaker_role(role):
		match = re.search(r'(andre |förste |tredje )?(vice )?talman', role)
		return match.group(0)
	db['role'] = db['role'].apply(_format_speaker_role)
	return db


class Corpus(pd.DataFrame):
	"""
	Store corpus metadata as a single pandas DataFrame where
	the column 'source' indicates the type of the row
	"""
	def __init__(self, *args, **kwargs):
		super(Corpus, self).__init__(*args, **kwargs)

	@property
	def _constructor(self):
		return Corpus

	def _load_metadata(self, file, source=False):
		df = pd.read_csv(f"corpus/metadata/{file}.csv")

		# Adjust to new structure where party information
		# is not included in member_of_parliament.csv
		if file == "member_of_parliament":
			print(df)
			columns = list(df.columns) + ["party"]
			party_df = pd.read_csv(f"corpus/metadata/party_affiliation.csv")
			party_df = party_df[party_df["start"].notnull()]
			party_df = party_df[party_df["end"].notnull()]
			df = df.merge(party_df, on=["wiki_id", "start", "end"], how="left")
			df = df[columns]
			print(df)
			print(df[df["party"].notnull()])
		if source:
			df['source'] = file
		return df

	def add_mps(self):
		df = self._load_metadata('member_of_parliament', source=True)
		df = infer_chamber(df)
		df = format_member_role(df)
		return Corpus(pd.concat([self, df]))
	        
	def add_ministers(self):
		df = self._load_metadata('minister', source=True)
		df = format_minister_role(df)
		return Corpus(pd.concat([self, df]))

	def add_speakers(self):
		df = self._load_metadata('speaker', source=True)
		df = infer_chamber(df)
		df = format_speaker_role(df)
		return Corpus(pd.concat([self, df]))

	def add_persons(self):
		df = self._load_metadata('person')
		return self.merge(df, on='wiki_id', how='left')

	def add_location_specifiers(self):
		df = self._load_metadata('location_specifier')
		return self.merge(df, on='wiki_id', how='left')

	def add_names(self):
		df = self._load_metadata('name')
		return self.merge(df, on='wiki_id', how='left')
	
	def impute_dates(self):
		return impute_date(self)

	def impute_parties(self):
		df = self._load_metadata('party_affiliation')
		df = impute_date(df)
		return impute_party(self, df)

	def abbreviate_parties(self):
		df = self._load_metadata('party_abbreviation')
		return abbreviate_party(self, df)

	def clean_names(self):
		return clean_name(self)




"""
Parla Clarin generation
"""
import pandas as pd
import progressbar, copy
from lxml import etree
from pyparlaclarin.create import pc_header, create_parlaclarin

from .utils import infer_metadata, get_formatted_uuid, XML_NS
from .download import get_blocks
from .db import year_iterator


def create_tei(root, metadata):
    """
    Create a Parla-Clarin TEI element from a list of segments.

    Args:
        txts: a list of lists of strings, corresponds to content blocks and paragraphs, respectively.
        metadata: Metadata of the parliamentary session
    """
    # TODO: Rewrite completely
    metadata = copy.deepcopy(metadata)

    tei = etree.Element("TEI")
    protocol_id = root.attrib["id"]
    metadata["document_title"] = (
        protocol_id.replace("_", " ").split("-")[0].replace("prot", "Protokoll")
    )
    documentHeader = pc_header(metadata)
    tei.append(documentHeader)

    text = etree.SubElement(tei, "text")
    front = etree.SubElement(text, "front")
    preface = etree.SubElement(front, "div", type="preface")
    etree.SubElement(preface, "head").text = protocol_id.split(".")[0]
    if "date" not in metadata:
        year = metadata.get("year", 2020)
        metadata["date"] = str(year) + "-01-01"

    etree.SubElement(preface, "docDate", when=metadata["date"]).text = metadata.get(
        "date", "2020-01-01"
    )

    body = etree.SubElement(text, "body")
    body_div = etree.SubElement(body, "div")

    current_speaker = None
    current_page = -1

    pb = etree.SubElement(body_div, "pb")
    pb.attrib["n"] = str(current_page)
    page_url = "https://betalab.kb.se/" + protocol_id + "/"
    page_filename = (
        protocol_id.replace("-", "_") + "-{:03d}".format(current_page) + ".jp2/_view"
    )
    pb.attrib["facs"] = page_url + page_filename

    element_seed = ""
    for content_block in root:
        new_page = content_block.attrib.get("page", current_page)
        new_page = int(new_page)
        if new_page != current_page:
            current_page = new_page
            element_seed = f"{protocol_id}\n{current_page}\n"
            pb = etree.SubElement(body_div, "pb")
            pb.attrib["n"] = str(current_page)
            page_url = "https://betalab.kb.se/" + protocol_id + "/"
            page_filename = (
                protocol_id.replace("-", "_")
                + "-{:03d}".format(current_page)
                + ".jp2/_view"
            )
            pb.attrib["facs"] = page_url + page_filename

        for textblock in content_block:
            text = " ".join(textblock.text.split())
            if text == "":
                continue
            note = etree.SubElement(body_div, "note")
            note.text = text

            # Generate reproducible UUID
            element_seed += text
            note.attrib[f"{XML_NS}id"] = get_formatted_uuid(element_seed)

    return tei

def dict_to_tei(data):
    metadata = copy.deepcopy(data)

    tei = etree.Element("TEI")
    protocol_id = metadata["protocol_id"]
    metadata["document_title"] = (
        protocol_id.replace("_", " ").split("-")[0].replace("prot", "Protokoll")
    )
    documentHeader = pc_header(metadata)
    tei.append(documentHeader)

    text = etree.SubElement(tei, "text")
    front = etree.SubElement(text, "front")
    preface = etree.SubElement(front, "div", type="preface")
    etree.SubElement(preface, "head").text = protocol_id.split(".")[0]
    if "date" not in metadata:
        year = metadata.get("year", 2020)
        metadata["date"] = str(year) + "-01-01"

    etree.SubElement(preface, "docDate", when=metadata["date"]).text = metadata.get(
        "date", "2020-01-01"
    )

    body = etree.SubElement(text, "body")
    body_div = etree.SubElement(body, "div")

    for paragraph in data["paragraphs"]:
        note = etree.SubElement(body_div, "note")
        note.text = paragraph

    return tei


def gen_parlaclarin_corpus(
    protocol_db,
    archive,
    corpus_metadata=dict(),
    str_output=True,
):
    """
    Create a parla-clarin file out of all protocols that are provided.
    """
    teis = []

    for ix, package in list(protocol_db.iterrows()):
        protocol_id = package["protocol_id"]
        pages = package["pages"]
        metadata = infer_metadata(protocol_id)
        protocol = get_blocks(protocol_id, archive)
        tei = create_tei(protocol, metadata)
        teis.append(tei)

    corpus_metadata["edition"] = "0.1.0"
    corpus = create_parlaclarin(teis, corpus_metadata)
    return corpus

def dict_to_parlaclarin(data):
    """
    Create per-protocol parlaclarin files of all files provided in file_db.
    """
    session = data["session"]
    default_metadata = dict(
        document_title=f"Riksdagens protocols {session}",
        authority="National Library of Sweden and the WESTAC project",
        correction="Some data curation was done. It is documented in input/curation/instances",
        edition="0.4.2",
    )
    for key in default_metadata:
        if key not in data:
            data[key] = default_metadata[key]

    protocol_id = data["protocol_id"]
    yearstr = protocol_id[5:]
    yearstr = yearstr.split("-")[0]
    parlaclarin_path = f"corpus/protocols/{yearstr}/{protocol_id}.xml"

    tei = dict_to_tei(data)
    parla_clarin_str = create_parlaclarin(tei, data)
    f = open(parlaclarin_path, "w")
    f.write(parla_clarin_str)
    f.close()


def parlaclarin_workflow_individual(file_db, archive, corpus_metadata=dict()):
    """
    Create per-protocol parlaclarin files of all files provided in file_db.
    """
    for corpus_year, package_ids, year_db in year_iterator(file_db):
        print("Generate corpus for year", corpus_year)

        default_metadata = dict(
            document_title="Riksdagens protocols " + str(corpus_year),
            authority="National Library of Sweden and the WESTAC project",
            correction="Some data curation was done. It is documented in input/curation/instances",
        )
        for key in default_metadata:
            if key not in corpus_metadata:
                corpus_metadata[key] = default_metadata[key]

        year_db = file_db[file_db["year"] == corpus_year]
        for ix, row in progressbar.progressbar(list(year_db.iterrows())):
            df = pd.DataFrame([row], columns=year_db.columns)
            protocol_id = row["protocol_id"]
            parla_clarin_str = gen_parlaclarin_corpus(
                df,
                archive,
                corpus_metadata=corpus_metadata,
            )

            yearstr = protocol_id[5:]
            yearstr = yearstr.split("-")[0]
            parlaclarin_path = "corpus/protocols/" + yearstr + "/" + protocol_id + ".xml"
            f = open(parlaclarin_path, "w")
            f.write(parla_clarin_str)
            f.close()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['pyriksdagen']

package_data = \
{'': ['*'], 'pyriksdagen': ['data/queries/*', 'data/queries_backup/*']}

install_requires = \
['SPARQLWrapper',
 'base58',
 'beautifulsoup4',
 'dateparser',
 'importlib_resources',
 'kblab-client',
 'lxml',
 'matplotlib',
 'numpy',
 'pandas',
 'progressbar2',
 'pyparlaclarin',
 'requests',
 'tqdm',
 'unidecode']

setup_kwargs = {
    'name': 'pyriksdagen',
    'version': '0.8.2rc3',
    'description': 'Access the Riksdagen corpus',
    'long_description': '[![Check unchanged](https://github.com/welfare-state-analytics/riksdagen-corpus/actions/workflows/check_unchanged.yml/badge.svg)](https://github.com/welfare-state-analytics/riksdagen-corpus/actions/workflows/check_unchanged.yml)\n[![Unit tests](https://github.com/welfare-state-analytics/riksdagen-corpus/actions/workflows/push.yml/badge.svg)](https://github.com/welfare-state-analytics/riksdagen-corpus/actions/workflows/push.yml)\n[![Validate Parla-Clarin XML](https://github.com/welfare-state-analytics/riksdagen-corpus/actions/workflows/validate.yml/badge.svg)](https://github.com/welfare-state-analytics/riksdagen-corpus/actions/workflows/validate.yml)\n\n# Swedish parliamentary proceedings - Riksdagens protokoll 1867-today\n\n_Westac Project, 2020-2023_\n\nThe full data set consists of multiple parts:\n\n- Riksdagens protokoll between from 1867 until today in the [Parla-clarin](https://github.com/clarin-eric/parla-clarin) format\n- Comprehensive list of MPs and cabinet members during this period\n- [Documentation](https://github.com/welfare-state-analytics/riksdagen-corpus/wiki/) of the corpus and the curation process\n- [A Google Colab notebook](https://colab.research.google.com/drive/1C3e2gwi9z83ikXbYXNPfB6RF7spTgzxA?usp=sharing) that demonstrates how the dataset can be used with Python\n\n## Basic use\n\nA full dataset is available under [this download link](https://github.com/welfare-state-analytics/riksdagen-corpus/releases/latest/download/corpus.zip). It has the following structure\n\n- Annual protocol files in the ```corpus/protocols/``` folder\n- Structured metadata on MPs, speakers, ministers, and governments in the ```corpus/metadata/``` folder\n\nThe workflow to use the data is demonstrated in [this Google Colab notebook](https://colab.research.google.com/drive/1C3e2gwi9z83ikXbYXNPfB6RF7spTgzxA?usp=sharing).\n\n## Design choices of the project\n\nThe Riksdagen corpus is released as an iterative process, where the corpus is curated and expanded. Semantic versioning is used for the whole corpus, following the established major-minor-patch practices as they apply to data. For each major and minor release, a statistical sample is drawn, annotated and quantitatively evaluated. Errors are fixed as they are detected in order of priority. Moreover, the edit history is kept as a traceable git repository.\n\n![Estimate of mapping accuracy](https://raw.githubusercontent.com/welfare-state-analytics/riksdagen-corpus/main/input/accuracy/version_plot.png)\n\nWhile the contents of the corpus will change due to curation and expansion, we aim to keep the deliverable API, the corpus/ folder, as stable as possible. This means we avoid relocating files or folders, changing formats, changing columns in metadata files, or any other changes that might break downstream scripts. Conversely, files outside the corpus/ folder are internal to the project. End users may find utility in them but we make no effort to keep them consistent.\n\nThe data in the corpus is delivered as TEI XML files to follow established practices. The metadata is delivered as CSV files, following a normal form database structure while allowing for a legible git history. A more detailed description of the data and metadata structure and formats can be found in the README files in the corpus/ folder.\n\n## Participate in the curation process\n\nIf you find any errors, it is possible to submit corrections to them. This is documented in the [project wiki](https://github.com/welfare-state-analytics/riksdagen-corpus/wiki/Submit-corrections).\n',
    'author': 'ninpnin',
    'author_email': 'vainoyrjanainen@icloud.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/welfare-state-analytics/riksdagen-corpus',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
